import string
from collections.abc import Callable
from pathlib import Path
from textwrap import dedent

import pytest

from testlib.fixtures.command import Command
from testlib.fixtures.file_helper import File, AssetSymlink, with_files
from testlib.fixtures.snapshot import Snapshot
from testlib.utils import get_functional2_lang_files


@pytest.mark.parametrize("pytest_command", [["-k", "generic_test", "--setup-plan"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {"generic_test": {"in.nix": File("{}"), "eval-okay.out.exp": File("{ }\n")}}
            }
        }
    )
)
def test_detects_generic_lang_test(pytest_command: Command):
    result = pytest_command.run().ok()
    assert "lang/test_lang.py::test_eval_okay[generic_test:eval-okay]" in result.stdout_plain


@pytest.mark.parametrize("pytest_command", [["-k", "toml_test", "--setup-plan"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml_test": {
                        "in.nix": File("{}"),
                        "my_name.out.exp": File("{ }\n"),
                        "test.toml": File(
                            dedent("""
                                    [[test]]
                                    name = "my_name"
                                    runner = "eval-okay"
                                    """)
                        ),
                    }
                }
            }
        }
    )
)
def test_detects_toml_lang_test(pytest_command: Command):
    result = pytest_command.run().ok()
    assert "lang/test_lang.py::test_eval_okay[toml_test:my_name]" in result.stdout_plain


@pytest.mark.parametrize("pytest_command", [["--setup-plan"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "some_py_module": {
                        "in.nix": File("{}"),
                        "eval-okay.out.exp": File("{ }\n"),
                        "__init__.py": File(""),
                    }
                }
            }
        }
    )
)
def test_skips_py_files(files: Path, pytest_command: Command):
    result = pytest_command.run().ok()
    assert "lang/test_lang.py::test_eval_okay[some_py_module:eval-okay]" not in result.stdout_plain
    assert (
        f"[    INFO] [lang-test-collector] skipping {files.absolute()}/functional2/lang/some_py_module as it contains a py file, assuming custom tests"
        in result.stdout_plain
    )


@pytest.mark.parametrize("pytest_command", [["-k", "n_suffix"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "n_suffix": {
                        "in-1.nix": File("{}"),
                        "in-2.nix": File("{}"),
                        "eval-okay-1.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "eval-okay-2.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                    }
                }
            }
        }
    )
)
def test_collects_with_n_suffix(pytest_command: Command):
    result = pytest_command.run().ok()
    out = result.stdout_plain
    assert "n_suffix:eval-okay-1" in out
    assert "n_suffix:eval-okay-2" in out


@pytest.mark.parametrize("pytest_command", [["-k", "short_string_suffix"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "short_string_suffix": {
                        "in-speaker.nix": File("{}"),
                        "in-microphone.nix": File("{}"),
                        "eval-okay-speaker.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "eval-okay-microphone.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                    }
                }
            }
        }
    )
)
def test_collects_short_string_suffix(pytest_command: Command):
    result = pytest_command.run().ok()
    out = result.stdout_plain
    assert "short_string_suffix:eval-okay-speaker" in out
    assert "short_string_suffix:eval-okay-microphone" in out


@pytest.mark.parametrize("pytest_command", [["-k", "dash_suffix"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "dash_suffix": {
                        "in-string-with-dash.nix": File("{}"),
                        "in-some-more--dashes.nix": File("{}"),
                        "eval-okay-string-with-dash.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "eval-okay-some-more--dashes.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                    }
                }
            }
        }
    )
)
def test_collects_string_suffix_with_dash(pytest_command: Command):
    result = pytest_command.run().ok()
    out = result.stdout_plain
    assert "dash_suffix:eval-okay-string-with-dash" in out
    assert "dash_suffix:eval-okay-some-more--dashes" in out
    assert "dash_suffix:eval-okay-" in out


@pytest.mark.parametrize("pytest_command", [["-k", "bad_naming"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "bad_naming": {
                        "in-&x.nix": File("{}"),
                        "in-.nix": File("{}"),
                        "eval-okay-&x.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "eval-okay-.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                    }
                }
            }
        }
    )
)
def test_collection_fails_with_bad_naming(pytest_command: Command):
    result = pytest_command.run().expect(1)
    err = result.stdout_plain
    assert "bad_naming:eval-okay" in err
    assert "incorrectly formatted test name: 'eval-okay-'" in err
    assert "incorrectly formatted test name: 'eval-okay-&x'" in err


@pytest.mark.parametrize("pytest_command", [["-k", "infra and runners"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "infra_okay_runners": {
                        "in.nix": File("{}"),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "parse-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_po.out.exp"
                        ),
                    },
                    "infra_fail_runners": {
                        "in.nix": File("{"),
                        "eval-fail.err.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_ef.err.exp"
                        ),
                        "parse-fail.err.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_pf.err.exp"
                        ),
                    },
                }
            }
        }
    )
)
def test_all_runners_work(pytest_command: Command):
    result = pytest_command.run().ok()
    assert "lang/test_lang.py::test_eval_okay[infra_okay_runners:eval-okay]" in result.stdout_plain
    assert (
        "lang/test_lang.py::test_parse_okay[infra_okay_runners:parse-okay]" in result.stdout_plain
    )
    assert "lang/test_lang.py::test_eval_fail[infra_fail_runners:eval-fail]" in result.stdout_plain
    assert (
        "lang/test_lang.py::test_parse_fail[infra_fail_runners:parse-fail]" in result.stdout_plain
    )


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "generic_bad": {
                        "in.nix": File("{}"),
                        "fee-foo.out.exp": File(""),
                        "hello-okay.out.exp": File(""),
                        "parse-fops.out.exp": File(""),
                    }
                }
            }
        }
    )
)
def test_generic_bad_runner_name(pytest_command: Command):
    result = pytest_command.run().expect(1)
    err = result.stdout_s
    assert "test_invalid_configuration" in err
    assert "invalid runner name:" in err
    assert "Invalid configuration for 'generic_bad:fee-foo'" in err
    assert "Invalid configuration for 'generic_bad:hello-okay'" in err
    assert "Invalid configuration for 'generic_bad:parse-fops'" in err


@pytest.mark.parametrize("pytest_command", [["-k", "toml_test"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml_test": {
                        "in.nix": File("{}"),
                        "my_name.out.exp": File("{ }\n"),
                        "test.toml": File(
                            dedent("""
                                    [[test]]
                                    name = "my_name"
                                    runner = "plushies"
                                    """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_bad_runner_name(pytest_command: Command):
    result = pytest_command.run().expect(1)
    err = result.stdout_plain
    assert "test_invalid_configuration" in err
    assert "invalid runner name:" in err
    assert "Invalid configuration for 'toml_test:my_name':" in err


@pytest.mark.parametrize("pytest_command", [["-k", "toml_test"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml_test": {
                        "in.nix": File("{}"),
                        "my_name.out.exp": File("{ }\n"),
                        "test.toml": File(
                            dedent("""
                                    [[test]]
                                    name = "my_name"
                                    runner = "eval-okay"
                                    cuddles = true
                                    """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_too_many_args(pytest_command: Command):
    result = pytest_command.run().expect(1)
    err = result.stdout_plain
    assert "test_invalid_configuration" in err
    assert "unexpected arguments: ['cuddles']" in err


@pytest.mark.parametrize("pytest_command", [["-k", "toml_test"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml_test": {
                        "in.nix": File("{}"),
                        "my_name.out.exp": File("{ }\n"),
                        "test.toml": File(
                            dedent("""
                                    invalid_test = "eval-okay"
                                    [[test]]
                                    name = "my_name"
                                    runner = "eval-okay"
                                    flags = 1
                                    [[test]]
                                    name = "second"
                                    runner = "eval-okay"
                                    extra-files = [false, true, true]
                                    """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_invalid_argument_types(pytest_command: Command):
    result = pytest_command.run().expect(1)
    err = result.stdout_plain
    assert "test_invalid_configuration" in err
    assert "unexpected key(s) ['invalid_test']; expected only 'test'" in err
    assert "invalid value type for 'flags'" in err
    assert "invalid value type for 'extra-files':" in err


@pytest.mark.parametrize("pytest_command", [["-k", "toml_test"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml_test": {
                        "in.nix": File("{}"),
                        "my_name.out.exp": File("{ }\n"),
                        "test.toml": File(
                            dedent("""
                                    [[test]]
                                    name = "my_name"
                                    runner = "eval-okay"
                                    cuddles = True
                                    """)
                        ),
                    }
                }
            }
        }
    )
)
def test_invalid_toml(pytest_command: Command):
    result = pytest_command.run().expect(1)
    err = result.stdout_plain
    assert "test_invalid_configuration" in err
    assert "couldn't parse toml" in err


@pytest.mark.parametrize(
    "pytest_command", [(["-k", "update_test", "--accept-tests"], False)], indirect=True
)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {"update_test": {"in.nix": File("{}"), "eval-okay.out.exp": File("old")}}
            },
            "out.exp": AssetSymlink("assets/test_lang_infra/runner_eo.out.exp"),
        }
    )
)
def test_updates_expected_output(
    files: Path, pytest_command: Command, snapshot: Callable[[str], Snapshot]
):
    assert (files / "functional2/lang/update_test/eval-okay.out.exp").read_text() == "old"
    pytest_command.run().ok()
    assert (
        snapshot("out.exp")
        == (files / "functional2/lang/update_test/eval-okay.out.exp").read_text()
    )


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "generic_unused": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                    }
                }
            }
        }
    )
)
def test_generic_throws_unused_files(pytest_command: Command):
    res = pytest_command.run().expect(1)
    out = res.stdout_plain
    assert "test_invalid_configuration[generic_unused-reasons0]" in out
    assert "the following files weren't referenced: {'in-1.nix'}" in out


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml_unused": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "eval-fail.err.exp": File(""),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                    [[test]]
                                    runner = "eval-okay"
                                    """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_throws_unused_files(
    pytest_command: Command, balanced_templater: type[string.Template]
):
    res = pytest_command.run().expect(1)
    out = res.stdout_plain
    assert "test_invalid_configuration[toml_unused-reasons0]" in out
    pattern = balanced_templater("the following files weren't referenced: {'@A@', '@B@'}")
    assert pattern.substitute(A="eval-fail.err.exp", B="in-1.nix") in out or pattern.substitute(
        B="eval-fail.err.exp", A="in-1.nix"
    )


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml_unused": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "eval-fail.err.exp": File(""),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                    [[test]]
                                    runner = "eval-okay"
                                    [[test]]
                                    runner = "bad-name"
                                    """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_no_unused_on_invalid_config(pytest_command: Command):
    res = pytest_command.run().expect(1)
    log = res.stdout_plain
    assert "test_invalid_configuration[toml_unused:bad-name-reasons0]" in log
    assert "invalid runner name:" in log
    assert "the following files weren't referenced" not in log


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "generic_unused": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "shrimpy-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                    }
                }
            }
        }
    )
)
def test_generic_no_unused_on_invalid_config(pytest_command: Command):
    res = pytest_command.run().expect(1)
    out = res.stdout_plain
    assert "test_invalid_configuration[generic_unused:shrimpy-okay-reasons0]" in out
    assert "invalid runner name:" in out
    assert "the following files weren't referenced" not in out


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "single-multi": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "eval-fail.err.exp": File(""),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        runner = "eval-okay"
                                        in = ["in.nix", "in-1.nix"]
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_single_only_str(pytest_command: Command):
    res = pytest_command.run().expect(1)
    err = res.stdout_plain
    assert "test_invalid_configuration[single-multi:eval-okay-reasons0]" in err
    assert "invalid type for 'in': ['in.nix', 'in-1.nix'], expected a string" in err


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "multi-single": {
                        "in.nix": File("{}"),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        matrix = true
                                        runner = "eval-okay"
                                        in = "in.nix"
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_matrix_only_list_str(pytest_command: Command):
    res = pytest_command.run().expect(1)
    err = res.stdout_plain
    assert "test_invalid_configuration[multi-single:eval-okay-reasons0]" in err
    assert "invalid type for 'in': 'in.nix', expected a list of string" in err


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "matrix-all": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "eval-okay-1.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        matrix = true
                                        runner = "eval-okay"
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_matrix_uses_all_files(pytest_command: Command):
    res = pytest_command.run().ok()
    out = res.stdout_plain
    assert "test_eval_okay[matrix-all:eval-okay] PASSED" in out
    assert "test_eval_okay[matrix-all:eval-okay-1] PASSED" in out


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "mixed-matrix": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "eval-okay-1.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "non-matrix-1.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        matrix = true
                                        runner = "eval-okay"
                                        [[test]]
                                        name = "non-matrix"
                                        in = "in-1.nix"
                                        runner = "eval-okay"
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_mixing_matrix_single(pytest_command: Command):
    res = pytest_command.run().ok()
    out = res.stdout_plain
    assert "test_eval_okay[mixed-matrix:eval-okay] PASSED" in out
    assert "test_eval_okay[mixed-matrix:eval-okay-1] PASSED" in out
    assert "test_eval_okay[mixed-matrix:non-matrix-1] PASSED" in out


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "non-unique": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        matrix = true
                                        runner = "eval-okay"
                                        [[test]]
                                        in = "in-1.nix"
                                        runner = "eval-okay"
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_non_unique_name(pytest_command: Command):
    res = pytest_command.run().expect(1)
    err = res.stdout_plain
    assert "test_invalid_configuration[non-unique:eval-okay-1-reasons0]" in err
    assert (
        "id 'non-unique:eval-okay-1' is not unique. Please set the 'name' attribute manually" in err
    )


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "subset": {
                        "in.nix": File("{}"),
                        "in-1.nix": File("{}"),
                        "in-hello.nix": File("{}"),
                        "eval-okay-1.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "matrix.out.exp": AssetSymlink("assets/test_lang_infra/runner_eo.out.exp"),
                        "matrix-hello.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        name = "matrix"
                                        matrix = true
                                        in = ["in.nix", "in-hello.nix"]
                                        runner = "eval-okay"
                                        [[test]]
                                        runner = "eval-okay"
                                        in = "in-1.nix"
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_matrix_subset(pytest_command: Command):
    res = pytest_command.run().ok()
    out = res.stdout_plain
    assert "test_eval_okay[subset:matrix] PASSED" in out
    assert "test_eval_okay[subset:matrix-hello] PASSED" in out
    assert "test_eval_okay[subset:matrix-1] PASSED" not in out
    assert "test_eval_okay[subset:eval-okay-1] PASSED" in out
    assert "test_eval_okay[subset:eval-okay] PASSED" not in out


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "in-naming": {
                        "in-some-name.nix": File("{}"),
                        "eval-okay-some-name.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        matrix = true
                                        in = ["in-some-name", "-some-name.nix", "some-name.nix"]
                                        runner = "eval-okay"
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_bad_in_naming(pytest_command: Command):
    res = pytest_command.run().expect(1)
    out = res.stdout_plain
    assert "test_invalid_configuration[in-naming:eval-okay-reasons0]" in out
    for msg in [
        "invalid in-file name 'in-some-name' at position 0 for 'in'",
        "invalid in-file name '-some-name.nix' at position 1 for 'in'",
        "invalid in-file name 'some-name.nix' at position 2 for 'in'",
    ]:
        assert msg in out


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "toml-missing": {
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        ),
                        "test.toml": File(
                            dedent("""
                                        [[test]]
                                        runner = "eval-okay"
                                        """)
                        ),
                    }
                }
            }
        }
    )
)
def test_toml_missing_in_file(pytest_command: Command):
    res = pytest_command.run().expect(1)
    out = res.stdout_plain
    assert "ERROR lang/test_lang.py::test_eval_okay[toml-missing:eval-okay] - FileNotF" in out


@pytest.mark.parametrize("pytest_command", [[]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "generic-missing": {
                        "eval-okay.out.exp": AssetSymlink(
                            "assets/test_lang_infra/runner_eo.out.exp"
                        )
                    }
                }
            }
        }
    )
)
def test_generic_missing_in_file(pytest_command: Command):
    res = pytest_command.run().expect(1)
    out = res.stdout_plain
    assert "ERROR lang/test_lang.py::test_eval_okay[generic-missing:eval-okay] - FileNotF" in out


@pytest.mark.parametrize("pytest_command", [["--setup-plan"]], indirect=True)
@with_files(
    get_functional2_lang_files(
        {
            "functional2": {
                "lang": {
                    "generic-duplicate": {
                        "eval-okay.out.exp": File("A"),
                        "eval-okay.err.exp": File("B"),
                        "in.nix": File(""),
                    }
                }
            }
        }
    )
)
def test_generic_no_duplicate_collection(pytest_command: Command):
    res = pytest_command.run().ok()
    assert (
        "test_invalid_configuration[name0-reasons0] SKIPPED" in res.stdout_plain
        or "test_invalid_configuration[NOTSET]" in res.stdout_plain
    )



import dataclasses
from textwrap import dedent, indent

from common import (
    cxx_literal,
    generate_file,
    load_data,
    get_argument_parser,
    get_experimental_features,
)


@dataclasses.dataclass
class Builtin:
    name: str
    documentation: str
    args: list[str]
    experimental_feature: str | None = None
    implementation: str = ""
    rename_in_global_scope: bool = True

    def __post_init__(self):
        self.implementation = self.implementation or f"prim_{self.name}"

    def generate_code(self, experimental_features: dict[str, str]) -> str:
        xf = experimental_features[self.experimental_feature]
        cond = (
            f"if (experimentalFeatureSettings.isEnabled({xf})) "
            if self.experimental_feature
            else ""
        )
        return dedent(f"""
        {cond}{{
            addPrimOp({{
                .name = {cxx_literal(("__" if self.rename_in_global_scope else "") + self.name)},
                .args = {cxx_literal(self.args)},
                .arity = {len(self.args)},
                .doc = {cxx_literal(self.documentation)},
                .fun = {self.implementation},
                .experimentalFeature = {xf},
            }});
        }}
        """)

    @property
    def docs(self) -> str:
        return dedent(f"""
            <dt id="builtins-{self.name}">
              <a href="#builtins-{self.name}"><code>{self.name} {
            " ".join([f"<var>{arg}</var>" for arg in self.args])
        }</code></a>
            </dt>
            <dd>

{indent(self.documentation, "    " * 3)}

            {
            f"This function is only available if the [{self.experimental_feature}](@docroot@/contributing/experimental-features.md#xp-feature-{self.experimental_feature}) experimental feature is enabled."
            if self.experimental_feature is not None
            else ""
        }
            </dd>

        """)


def main():
    ap = get_argument_parser()
    ap.add_argument(
        "--experimental-features", help="Directory containing the experimental feature definitions"
    )
    args = ap.parse_args()

    builtins = load_data(args.defs, Builtin)

    experimental_features = get_experimental_features(
        args.experimental_features, [b.experimental_feature for (_, b) in builtins]
    )

    generate_file(
        args.header,
        builtins,
        lambda builtin: builtin.name,
        lambda b: b.generate_code(experimental_features),
    )
    generate_file(args.docs, builtins, lambda builtin: builtin.name, lambda b: b.docs)


if __name__ == "__main__":
    main()



import sys

import pytest

from testlib.fixtures.file_helper import with_files, CopyFile
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset

_mo_files = {"config.nix": get_global_asset("config.nix"), "drv.nix": CopyFile("assets/drv.nix")}


@pytest.fixture(autouse=True)
def commands(nix: Nix):
    nix.settings.add_xp_feature("nix-command")


@pytest.mark.skipif(
    sys.platform == "darwin",
    reason="installables from symlinks to darwin-fake-redirected store paths are broken",
)
@with_files(_mo_files)
def test_store_delete_unlink(nix: Nix):
    cmd = nix.nix(["build", "-f", "drv.nix", "-o", "result", "--print-out-paths"])
    cwd = cmd.cwd

    built = cmd.run().ok().stdout_plain
    built_path = nix.physical_store_path_for(built)
    assert built_path.exists()

    result_path = cwd.joinpath("result")
    assert result_path.exists(follow_symlinks=False)
    assert result_path.is_symlink()

    # Deleting the "symlink-to-store-path" installable should fail, because the symlink itself
    # keeps the store path alive.
    nix.nix(["store", "delete", "./result"]).run().expect(1)
    assert result_path.exists(follow_symlinks=False)
    assert built_path.exists()

    # But with --unlink, it'll remove that root and thus actually delete the path.
    nix.nix(["store", "delete", "--unlink", "./result"]).run().ok()
    assert not result_path.exists(follow_symlinks=False), "--unlink didn't remove gc root"
    assert not built_path.exists(), "nix store delete didn't actually delete"


@pytest.mark.skipif(
    sys.platform == "darwin",
    reason="installables from symlinks to darwin-fake-redirected store paths are broken",
)
@with_files(_mo_files)
def test_store_delete_unlink_closure(nix: Nix):
    cmd = nix.nix(["build", "-f", "drv.nix", "-o", "result", "--print-out-paths"])
    cwd = cmd.cwd

    built = cmd.run().ok().stdout_plain
    result_path = cwd.joinpath("result")

    # We just built some paths; let's make sure they exist now, before we delete them.
    requisites = nix.nix(["path-info", "--recursive", built]).run().ok().stdout_plain.splitlines()
    req_paths = [nix.physical_store_path_for(req) for req in requisites]
    for req in req_paths:
        assert req.exists()

    nix.nix(["store", "delete", "--unlink", "--delete-closure", "./result"]).run().ok()
    assert not result_path.exists(follow_symlinks=False)

    for req in req_paths:
        assert not req.exists(), f"{req} should be deleted but it's still here"



from pathlib import Path
from typing import Any

import pytest

from testlib.fixtures.file_helper import with_files, CopyFile, Symlink, File
from testlib.fixtures.nix import Nix


@pytest.fixture(autouse=True)
def nix_command_feature(nix: Nix):
    nix.settings.add_xp_feature("nix-command")


@pytest.mark.parametrize(
    ("stdin", "exp", "flags"),
    [
        (
            b"""
                {
                    bar = 3 + 1;
                    foo = 2 + 2;
                }
            """,
            "{ bar = 4; foo = 4; }",
            ["-f", "-"],
        )
    ],
)
@with_files({"eval.nix": CopyFile("assets/eval.nix")})
def test_valid_eval_stdin(nix: Nix, stdin: bytes, exp: Any, flags: list[str]):
    cmd = nix.nix(["eval", *flags]).with_stdin(stdin)
    res = cmd.run().ok()
    assert res.stdout_plain == exp


@with_files({"eval.nix": CopyFile("assets/eval.nix")})
def test_valid_filename_stdin(nix: Nix, files: Path):
    stdin = (files / "eval.nix").read_bytes()
    res = nix.nix(["eval", "int", "-f", "-"]).with_stdin(stdin).run().ok()
    assert res.stdout_plain == "123"

    res = nix.nix_instantiate(["-A", "int", "--eval", "-"]).with_stdin(stdin).run().ok()
    assert res.stdout_plain == "123"


@pytest.mark.parametrize(
    ("flags", "exp"),
    [
        (["int"], "123"),
        (["str"], '"foo\\nbar"'),
        (["str", "--raw"], "foo\nbar"),
        (["attr"], '{ foo = "bar"; }'),
        (["attr", "--json"], '{"foo":"bar"}'),
    ],
)
@with_files({"eval.nix": CopyFile("assets/eval.nix")})
def test_valid_eval_f(nix: Nix, flags: list[str], exp: str):
    # nix3 cli
    res = nix.nix(["eval", *flags, "-f", "./eval.nix"]).run().ok()
    assert res.stdout_plain == exp

    # legacy cli
    res = nix.nix_instantiate(["-A", *flags, "--eval", "./eval.nix"]).run().ok()
    assert res.stdout_plain == exp


@pytest.mark.parametrize(
    ("expr", "exp"),
    [("assert 1 + 2 == 3; true", "true"), ('{"assert"=1;bar=2;}', '{ "assert" = 1; bar = 2; }')],
)
@pytest.mark.parametrize("long", [False, True])
def test_valid_expr(nix: Nix, expr: str, exp: str, long: bool):
    flag = "--expr" if long else "-E"
    res = nix.nix(["eval", flag, expr]).run().ok()
    assert res.stdout_plain == exp

    res = nix.nix_instantiate(["--eval", flag, expr]).run().ok()
    assert res.stdout_plain == exp


@with_files({"eval.nix": CopyFile("assets/eval.nix")})
def test_invalid_no_coercible_values(nix: Nix):
    """Non-coercible values throws errors under `--raw`"""
    res = nix.nix(["eval", "int", "--raw", "-f", "./eval.nix"]).run().expect(1)
    assert "error: cannot coerce an integer to a string: 123" in res.stderr_plain

    res = nix.nix_instantiate(["-A", "int", "--raw", "./eval.nix"]).run().expect(1)
    assert (
        "error: expression was expected to be a derivation or collection of derivations, but instead was an integer"
        in res.stderr_plain
    )


def test_invalid_top_level_error(nix: Nix):
    """Top-level eval errors should be printed to stderr with a traceback."""
    res = nix.nix(["eval", "--expr", 'throw "a sample throw message"']).run().expect(1)
    assert "a sample throw message" in res.stderr_plain
    assert "caused by explicit throw" in res.stderr_plain


def test_valid_nested_error(nix: Nix):
    """But errors inside something should print an elided version, and exit with 0."""
    res = nix.nix(["eval", "--expr", '{ throws = throw "a sample throw message"; }']).run().ok()
    assert res.stdout_plain == "{ throws = «error: a sample throw message»; }"


def test_valid_restricted_toFile(nix: Nix):  # noqa: N802 # builtin name
    """Check if toFile can be utilized during restricted eval"""
    res = (
        nix.nix(["eval", "--restrict-eval", "--expr", 'import (builtins.toFile "source" "42")'])
        .run()
        .ok()
    )
    assert res.stdout_plain == "42"


@with_files({"cycle.nix": Symlink("cycle.nix")})
# timeout given in seconds
@pytest.mark.timeout(30)
def test_invalid_no_hang_symlink_cycle(nix: Nix):
    """Check that symlink cycles don't cause a hang."""
    res = nix.nix(["eval", "--file", "cycle.nix"]).run().expect(1)
    assert (
        "error: too many symbolic links encountered while traversing the path" in res.stderr_plain
    )


@with_files({"xyzzy": {"default.nix": File("123")}, "foo": {"bar": Symlink("../xyzzy")}})
def test_valid_symlink_resolution(nix: Nix):
    """Check that relative symlinks are resolved correctly."""
    res = nix.nix(["eval", "--impure", "--expr", "import ./foo/bar"]).run().ok()
    assert res.stdout_plain == "123"


def test_valid_warn_unknown_setting(nix: Nix):
    """Test that unknown settings are warned about"""
    res = (
        nix.nix(["eval", "--option", "foobar", "baz", "--expr", '"foxes are cute"', "--raw"])
        .run()
        .ok()
    )
    assert res.stdout_plain == "foxes are cute"
    assert "warning: unknown setting 'foobar'" in res.stderr_plain



from testlib.fixtures.file_helper import with_files, File
from testlib.fixtures.nix import Nix


@with_files({"foo": {"bar": File("bla")}})
def test_path_read_file(nix: Nix):
    nix.settings.add_xp_feature("nix-command", "flakes")
    res = (
        nix.nix(
            [
                "eval",
                "--raw",
                "--impure",
                "--expr",
                f'builtins.readFile (builtins.toString (builtins.fetchTree {{ type = "path"; path = "{nix.env.dirs.home / "foo"}"; }} + "/bar"))',
            ]
        )
        .run()
        .ok()
    )
    assert res.stdout_plain == "bla"


@with_files({"foo": {"bar": File("bla")}})
def test_path_read_dir(nix: Nix):
    nix.settings.add_xp_feature("nix-command", "flakes")
    res = (
        nix.nix(
            [
                "eval",
                "--json",
                "--impure",
                "--expr",
                f'builtins.readDir (builtins.toString (builtins.fetchTree {{ type = "path"; path = "{nix.env.dirs.home / "foo"}"; }}))',
            ]
        )
        .run()
        .ok()
    )
    assert res.stdout_plain == '{"bar":"regular"}'



from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import with_files
from testlib.utils import get_global_asset, CopyTemplate, CopyFile
from testlib.environ import environ
import pytest


@with_files(
    {
        "shell-hello.nix": CopyFile("assets/shell-hello.nix"),
        "config.nix": get_global_asset("config.nix"),
        "flake.nix": CopyTemplate("assets/run/flake.nix", {"system": environ.get("system")}),
    }
)
class TestFlakeRun:
    @pytest.mark.parametrize("attr", ["appAsApp", "pkgAsPkg"])
    def test_run_succeeds(self, nix: Nix, attr: str):
        assert (
            "Hello World"
            in nix.nix(["run", "--no-write-lock-file", f".#{attr}"], flake=True).run().ok().stdout_s
        )

    def test_run_pkg_as_app(self, nix: Nix):
        assert (
            "should have type 'derivation'"
            in nix.nix(["run", "--no-write-lock-file", ".#pkgAsApp"], flake=True)
            .run()
            .expect(1)
            .stderr_s
        )

    def test_run_app_as_pkg(self, nix: Nix):
        assert (
            "should have type 'app'"
            in nix.nix(["run", "--no-write-lock-file", ".#appAsPkg"], flake=True)
            .run()
            .expect(1)
            .stderr_s
        )



import subprocess
import json
from . import environment


def get_ephemeral_key(
        env: environment.RelengEnvironment) -> environment.S3Credentials:
    output = None
    command = [
        'garage-ephemeral-key', 'new', '--name', f'releng-{env.name}', '--read',
        '--write', '--age-secs', '3600',
        env.releases_bucket.removeprefix('s3://'),
        env.cache_bucket.removeprefix('s3://'),
        env.docs_bucket.removeprefix('s3://'),
    ]
    if env.s3_ssh_host is not None:
        command = ['ssh', f'{env.s3_ssh_user}@{env.s3_ssh_host}', *command]
    output = subprocess.check_output(command)
    d = json.loads(output.decode())
    return environment.S3Credentials(name=d['name'],
                                     id=d['id'],
                                     secret_key=d['secret_key'])



from pathlib import Path
from textwrap import dedent

import pytest

from testlib.fixtures.file_helper import File, with_files, EnvTemplate
from testlib.fixtures.nix import Nix

files = {
    "dep": {"flake.nix": File("{outputs = i: { };}")},
    "foo": {
        "flake.nix": EnvTemplate(
            dedent("""
            {
                inputs.a.url = "path:@HOME@/dep";

                outputs = i: {
                    sampleOutput = 1;
                };
            }
        """)
        )
    },
    "bar": {
        "flake.nix": EnvTemplate(
            dedent("""
            {
                inputs.b.url = "path:@HOME@/dep";

                outputs = i: {
                    sampleOutput = 1;
                };
            }
        """)
        )
    },
    "err": {"flake.nix": File('throw "error"')},
}


@pytest.fixture(autouse=True)
def nix_command_feature(nix: Nix):
    # "flakes" seems to be the feature actually responsible for
    # resolving the completions correctly
    #
    # rootile: no clue why...
    nix.settings.add_xp_feature("nix-command", "flakes")


def test_completions_valid(nix: Nix):
    nix.env.set_env("NIX_GET_COMPLETIONS", "1")
    res = nix.nix(["buil"]).run().ok()
    assert res.stdout_s == "normal\nbuild\t\n"

    nix.env.set_env("NIX_GET_COMPLETIONS", "2")
    res = nix.nix(["flake", "metad"]).run().ok()
    assert res.stdout_s == "normal\nmetadata\t\n"


def test_completions_invalid_envar(nix: Nix):
    nix.env.set_env("NIX_GET_COMPLETIONS", "-")
    res = nix.nix([]).run().expect(1)
    assert (
        "error: Invalid value for environment variable NIX_GET_COMPLETIONS: -\n" in res.stderr_plain
    )


@pytest.mark.parametrize("position", [0, 4])
def test_completions_invalid_position(nix: Nix, position: int):
    nix.env.set_env("NIX_GET_COMPLETIONS", str(position))
    res = nix.nix(["build", "a"]).run().expect(1)
    assert f"error: Invalid word number to get completion for: {position}" in res.stderr_plain


@with_files({"foo": {}})
def test_completions_existing_filename(nix: Nix):
    nix.env.set_env("NIX_GET_COMPLETIONS", "2")
    res = nix.nix(["build", "./f"]).run().ok()
    assert res.stdout_s == "filenames\n./foo\t\n"


def test_completions_nonexisting_filename(nix: Nix):
    nix.env.set_env("NIX_GET_COMPLETIONS", "2")
    res = nix.nix(["build", "./nonexistent"]).run().ok()
    assert res.stdout_s == "filenames\n"


@pytest.mark.parametrize(
    ("args", "pos"),
    [
        (["build", "./foo", "--override-input", ""], 4),
        (["flake", "show", "./foo", "--override-input", ""], 5),
        (["flake", "update", "--flake", "./foo", ""], 5),
        # tilde expansion
        (["build", "~/foo", "--override-input", ""], 4),
        (["flake", "update", "--flake", "~/foo", ""], 5),
        # Out of order
        (["build", "--override-input", "", "", "./foo"], 3),
    ],
)
@with_files(files)
def test_completions_input_override(nix: Nix, args: list[str], pos: int):
    nix.env.set_env("NIX_GET_COMPLETIONS", str(pos))
    res = nix.nix(args).run().ok()
    assert res.stdout_s == "normal\na\t\n"


@pytest.mark.parametrize(
    ("args", "pos"),
    [
        (["build", "./foo", "./bar", "--override-input", ""], 5),
        (["build", "./foo", "--override-input", "", "", "./bar"], 4),
    ],
)
@with_files(files)
def test_completions_multiple_input_flakes(nix: Nix, args: list[str], pos: int):
    nix.env.set_env("NIX_GET_COMPLETIONS", f"{pos}")
    res = nix.nix(args).run().ok()
    assert res.stdout_s == "normal\na\t\nb\t\n"


@with_files(files)
def test_completions_flake_update(nix: Nix, files: Path):
    cwd = files / "foo"
    nix.env.set_env("NIX_GET_COMPLETIONS", "3")
    cmd = nix.nix(["flake", "update", ""])
    cmd.cwd = cwd
    res = cmd.run().ok()
    assert res.stdout_s == "normal\na\t\n"


def test_flag_completion(nix: Nix):
    nix.env["NIX_GET_COMPLETIONS"] = "2"
    res = nix.nix(["build", "--dry"]).run().ok()
    assert "--dry-run" in res.stdout_plain
    assert "Show what this command would do without doing it" in res.stdout_plain


def test_option_completion(nix: Nix):
    nix.env["NIX_GET_COMPLETIONS"] = "3"
    res = nix.nix(["build", "--option", "allow-import-"]).run().ok()
    assert "allow-import-from-derivation" in res.stdout_plain


@pytest.mark.skip("SKIP(Hufschmitt 2022-07): Options as CLI-flags currently don't have completions")
def test_option_cli_flag_completion(nix: Nix):
    nix.env["NIX_GET_COMPLETIONS"] = "2"
    res = nix.nix(["build", "--allow-import-from"]).run().ok()
    assert "allow-import-from-derivation" in res.stdout_plain


@pytest.mark.parametrize(
    ("args", "pos", "exp"),
    [
        (["./foo#sam"], 2, "attrs\n./foo#sampleOutput\t\n"),
        (["./err#"], 2, "attrs\n"),
        (["--file", "./foo/flake.nix", "outp"], 4, "attrs\noutputs\t\n"),
        (["--file", "./err/flake.nix", "outp"], 4, "attrs\n"),
    ],
)
@with_files(files)
def test_valid_attr_path_completion(nix: Nix, args: list[str], pos: int, exp: str):
    nix.env["NIX_GET_COMPLETIONS"] = f"{pos}"
    res = nix.nix(["eval", *args]).run().ok()
    assert res.stdout_s == exp



from testlib.fixtures.file_helper import with_files, CopyFile
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack

import pytest

_files = {
    "check-refs.nix": CopyFile("assets/test_reference_checks/check-refs.nix"),
    **get_global_asset_pack("dependencies"),
}


@with_files(_files)
def test_references_detected(nix: Nix):
    dep, test1, test2 = (
        nix.nix_build(["check-refs.nix", "-A", "dep", "-A", "test1", "-A", "test2"])
        .run()
        .ok()
        .stdout_plain.splitlines()
    )

    refs_test1, refs_test2 = (
        nix.nix_store(["-q", "--references", path]).run().ok().stdout_plain.splitlines()
        for path in [test1, test2]
    )

    assert dep in refs_test1
    assert test1 not in refs_test1

    assert dep not in refs_test2
    assert test2 not in refs_test2

    # test2 has a reference to file `aux-ref` (${src})
    assert any("aux-ref" in p for p in refs_test2)


@with_files(_files)
def test_allowed_references_error(nix: Nix):
    # empty `allowedReferences` throws on detected references
    error = nix.nix_build(["check-refs.nix", "-A", "test3"]).run().expect(1).stderr_plain
    assert "check-refs-3' is not allowed to have direct references to the following paths:" in error


@with_files(_files)
@pytest.mark.parametrize("ok_attr", ["test4", "test5", "test7"])
def test_allowed_references_ok(nix: Nix, ok_attr: str):
    # two derivations where the path matches allowedReferences
    nix.nix_build(["check-refs.nix", "-A", ok_attr]).run().ok()


@with_files(_files)
def test_allowed_reference_self_reference(nix: Nix):
    # self-reference that's not allowed
    error = (
        nix.nix_build(["check-refs.nix", "-A", "test6"]).run().expect(1).stderr_plain.splitlines()
    )
    assert (
        "check-refs-6' is not allowed to have direct references to the following paths:"
        in error[-2]
    )
    assert "check-refs-6" in error[-1]


@with_files(_files)
def test_tofile_no_drv_output(nix: Nix):
    error = nix.nix_build(["check-refs.nix", "-A", "test8"]).run().expect(1).stderr_plain

    assert (
        "error: files created by builtins.toFile may not reference derivations, but builder.sh references"
        in error
    )


@with_files(_files)
def test_disallowed_references(nix: Nix):
    # disallowedReferences error fires
    error = (
        nix.nix_build(["check-refs.nix", "-A", "test9"]).run().expect(1).stderr_plain.splitlines()
    )

    assert (
        "check-refs-9' is not allowed to have direct references to the following paths" in error[-2]
    )
    assert "dependencies-top" in error[-1]

    # Dependencies from `disallowedReferences` are used at build-time, but are not
    # referenced by the final store path -> success.
    nix.nix_build(["check-refs.nix", "-A", "test10"]).run().ok()


@with_files(_files)
def test_structured_attrs_discard(nix: Nix):
    result = nix.nix_build(["check-refs.nix", "-A", "test11"]).run().ok().stdout_plain.splitlines()

    assert not (nix.nix_store(["-q", "--references", *result]).run().ok().stdout_plain.splitlines())


@with_files(_files)
def test_reference_specifier(nix: Nix):
    error = (
        nix.nix_build(["check-refs.nix", "-A", "test12"]).run().expect(1).stderr_plain.splitlines()
    )

    assert "output check for 'lib' contains an illegal reference specifier 'dev'" in error[-1]


@with_files(
    get_global_asset_pack("dependencies")
    | {
        "regression-reference-checks.nix": CopyFile(
            "assets/test_reference_checks/regression-reference-checks.nix"
        )
    }
)
def test_regression_partial_build(nix: Nix):
    out, _ = (
        nix.nix_build(
            ["regression-reference-checks.nix", "-A", "out", "-A", "man", "--no-out-link"]
        )
        .run()
        .ok()
        .stdout_plain.splitlines()
    )

    nix.nix_store(["--delete", out]).run().ok()

    nix.nix_build(["regression-reference-checks.nix", "-A", "out"]).run().ok()



import json
from collections.abc import Callable
from logging import Logger
from pathlib import Path

import pytest
import yaml
from _pytest.python import Metafunc

from lang.lang_util import LangTest, fetch_all_lang_tests, LangTestRunner
from testlib.fixtures.nix import Nix, with_diverted_store
from testlib.fixtures.snapshot import Snapshot


def pytest_generate_tests(metafunc: Metafunc):
    """
    This hook parametrizes the parser and eval tests with all test cases found in functional2/lang
    :param metafunc: the test function to parametrize, provided by pytest
    """
    func_name = metafunc.function.__name__
    tests, invalid_tests = fetch_all_lang_tests()
    if func_name == "test_invalid_configuration":
        metafunc.parametrize(("name", "reasons"), [(t.name, t.reasons) for t in invalid_tests])
        return
    selected_runner: LangTestRunner
    match func_name:
        case "test_eval_okay":
            selected_runner = LangTestRunner.EVAL_OKAY
        case "test_eval_fail":
            selected_runner = LangTestRunner.EVAL_FAIL
        case "test_parse_okay":
            selected_runner = LangTestRunner.PARSE_OKAY
        case "test_parse_fail":
            selected_runner = LangTestRunner.PARSE_FAIL
        case _:
            return
    selected_tests = tests[selected_runner]
    if len(selected_tests) > 0:
        # ignoring type here, because map doesn't recognize the typing correctly
        # due to returning multiple things and it expecting a single generic
        files, flags, ids = zip(*map(LangTest.to_params, selected_tests))  # type: ignore
    else:
        files, flags, ids = [], [], []
    metafunc.parametrize(("files", "flags"), zip(files, flags), ids=ids, indirect=["files"])


def _cleanup_output(stdout: str, stderr: str, origin: Path) -> tuple[str, str]:
    """
    Cleans up any paths present in stdout and stderr by replacing them with placeholders
    """
    test_path = str(origin)
    clean_out = stdout.replace(test_path, "/pwd")
    clean_err = stderr.replace(test_path, "/pwd")
    return clean_out, clean_err


def test_parse_okay(files: Path, nix: Nix, flags: list[str], snapshot: Callable[[str], Snapshot]):
    nix_command = nix.nix_instantiate(
        ["--parse", *flags, files / "in.nix"],
        # TODO(rootile): Mirrors behavior of init.sh from functional
        #  keep this for migration, but make it declarative afterwards
        #  and only active for the tests which need it
        flake=True,
    )
    result = nix_command.run().ok()
    stdout, stderr = _cleanup_output(result.stdout_s, result.stderr_s, files)

    # Taken from https://stackoverflow.com/a/39681672
    # Pyyaml does not give list items extra indentation, unlike pretty much everyone else.
    # For migration compatibility with the `yq` output from the functional/lang tests, we
    # use a custom dumper that produces the correct indentation.
    # Upstream issue (open since 2018): https://github.com/yaml/pyyaml/issues/234
    class CustomFixedIndentationDumper(yaml.Dumper):
        def increase_indent(self, flow: bool = False, indentless: bool = False):  # noqa: ARG002
            super().increase_indent(flow, False)

    result_obj = json.loads(stdout)
    result_yaml = yaml.dump(
        result_obj, Dumper=CustomFixedIndentationDumper, default_flow_style=False
    )
    # parser out are in a yaml format, which is why we convert it here for better viewing and editing
    assert snapshot("out.exp") == result_yaml
    assert snapshot("err.exp") == stderr


def test_parse_fail(files: Path, nix: Nix, flags: list[str], snapshot: Callable[[str], Snapshot]):
    nix_command = nix.nix_instantiate(["--parse", *flags, files / "in.nix"], flake=True)
    result = nix_command.run().expect(1)
    stdout, stderr = _cleanup_output(result.stdout_s, result.stderr_s, files)

    assert snapshot("out.exp") == stdout
    assert snapshot("err.exp") == stderr


@with_diverted_store
def test_eval_okay(files: Path, nix: Nix, flags: list[str], snapshot: Callable[[str], Snapshot]):
    nix_command = nix.nix_instantiate(["--eval", "--strict", *flags, files / "in.nix"], flake=True)
    result = nix_command.run().ok()
    stdout, stderr = _cleanup_output(result.stdout_s, result.stderr_s, files)

    assert snapshot("out.exp") == stdout
    assert snapshot("err.exp") == stderr


@with_diverted_store
def test_eval_fail(files: Path, nix: Nix, flags: list[str], snapshot: Callable[[str], Snapshot]):
    nix_command = nix.nix_instantiate(
        ["--eval", "--strict", "--show-trace", *flags, files / "in.nix"], flake=True
    )
    result = nix_command.run().expect(1)
    stdout, stderr = _cleanup_output(result.stdout_s, result.stderr_s, files)

    assert snapshot("out.exp") == stdout
    assert snapshot("err.exp") == stderr


def test_invalid_configuration(name: str, reasons: list[str], logger: Logger):
    msg = f"Invalid configuration for {name!r}: {reasons}"
    logger.error(msg)
    pytest.fail(msg)



from testlib.fixtures.file_helper import with_files, CopyFile
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset

_files = {
    "timeout.nix": CopyFile("assets/test_timeout/timeout.nix"),
    "config.nix": get_global_asset("config.nix"),
}


@with_files(_files)
def test_timeout_timeout(nix: Nix):
    res = (
        nix.nix_build(["-Q", "timeout.nix", "-A", "infiniteLoop", "--timeout", "2"])
        .run()
        .expect(101)
    )
    assert "timed out" in res.stderr_plain


@with_files(_files)
def test_timeout_max_log(nix: Nix):
    res = (
        nix.nix_build(["-Q", "timeout.nix", "-A", "infiniteLoop", "--max-build-log-size", "100"])
        .run()
        .expect(1)
    )
    assert "killed after writing more than 100 bytes of log output" in res.stderr_plain


@with_files(_files)
def test_timeout_silent(nix: Nix):
    res = nix.nix_build(["timeout.nix", "-A", "silent", "--max-silent-time", "2"]).run().expect(101)
    assert "file timed out after 2 seconds of silence" in res.stderr_plain


@with_files(_files)
def test_timeout_close_log(nix: Nix):
    res = nix.nix_build(["timeout.nix", "-A", "closeLog"]).run().expect(100)
    assert "failed due to signal 9 (Killed" in res.stderr_plain


@with_files(_files)
def test_timeout_nix3_silent(nix: Nix):
    res = (
        nix.nix(["build", "-f", "timeout.nix", "silent", "--max-silent-time", "2"], flake=True)
        .run()
        .expect(1)
    )
    assert "from .drv file timed out after 2 seconds of silence" in res.stderr_plain



from testlib.fixtures.nix import Nix


evil_drv = """
  builtins.derivation {
    name = "foo";
    builder = "/bin/sh";
    system = builtins.currentSystem;
    __structuredAttrs = true;

    __json.hello = "1";
    env.hello = "2";
  }
"""


def test_improper_structured_attrs_drv(nix: Nix):
    result = nix.nix_build(["--expr", evil_drv]).run().expect(1)
    assert "a `__json` attribute cannot be passed" in result.stderr_s



import xattr
import pytest

from pathlib import Path
from testlib.fixtures.env import ManagedEnv


def has_xattrs(path: Path) -> bool:
    """
    Check if a file or directory has any extended attributes.
    """
    try:
        xattrs_list = xattr.list(str(path), nofollow=True)
        return bool(xattrs_list)
    except OSError:
        # If an error occurs, either because the file doesn't exist or no xattrs, return False
        return False


def verify_no_xattrs_in_tree(root_dir: Path) -> None:
    """
    Traverse a directory tree and verify no file or directory has xattrs.
    """
    if root_dir.is_file():
        assert not has_xattrs(root_dir)
        return

    for entry in root_dir.iterdir():
        assert not has_xattrs(entry)
        if entry.is_dir():
            verify_no_xattrs_in_tree(entry)


def test_set_clear_xattrs_in(dir_: Path) -> bool:
    try:
        xattr.set(dir_, "user.test", "1", nofollow=True)
        xattr.remove(dir_, "user.test", nofollow=True)
        return True
    except OSError:
        return False


def skip_if_xattrs_are_unsupported(env: ManagedEnv) -> None:
    """
    Skip the current test if xattrs are unsupported in, either:
    - the test root
    - the fixture root
    """
    if not test_set_clear_xattrs_in(env.dirs.test_root):
        pytest.skip("xattrs cannot be used in the test root")

    if not test_set_clear_xattrs_in(env.dirs.home):
        pytest.skip("xattrs cannot be used in the fixture directory")



from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset
from testlib.fixtures.file_helper import File, with_files


@with_files(
    {
        "config.nix": get_global_asset("config.nix"),
        "default.nix": File("""
            with import ./config.nix;

            mkDerivation {
              name = "pass-as-file";
              passAsFile = [ "foo" ];
              foo = [ "xyzzy" ];
              builder = builtins.toFile "builder.sh" ''
                [ "$(basename $fooPath)" = .attr-1bp7cri8hplaz6hbz0v4f0nl44rl84q1sg25kgwqzipzd1mv89ic ]
                [ "$(cat $fooPath)" = xyzzy ]
                touch $out
              '';
            }
        """),
    }
)
def test_pass_as_file(nix: Nix):
    nix.nix_build([]).run().ok()



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File
from testlib.environ import environ
from pathlib import Path
import pytest
from .common import simple_flake

system = environ.get("system")

files = simple_flake() | {"foo": {"flake.nix": File("{ outputs = _: {}; }")}, "subdir": {}}


@pytest.fixture(autouse=True)
def common_init(nix: Nix, files: Path):
    nix.settings.add_xp_feature("nix-command", "flakes")

    flake = files / "flake.nix"
    flake.write_text(f"""
        {{
            inputs.foo.url = "{files}/foo";
            outputs = a: {{
               packages.{system} = rec {{
                 test = import ./simple.nix;
                 default = test;
               }};
            }};
        }}
    """)


good_uris: list[list[str]] = [
    [],
    ["."],
    [".#"],
    [".#test"],
    ["../subdir"],
    ["../subdir#test"],
    ["$PWD"],
]


@with_files(files)
class TestFlakeSearchPath:
    @pytest.mark.parametrize("args", good_uris)
    def test_flake_search_goes_up(self, nix: Nix, files: Path, args: list[str]):
        subdir = files / "subdir"
        nix.nix(
            ["build", *[arg.replace("$PWD", str(subdir)) for arg in args]], cwd=subdir
        ).run().ok()

    def test_flake_search_inactive_with_path_uri(self, nix: Nix, files: Path):
        subdir = files / "subdir"
        assert (
            "does not contain a '/flake.nix' file"
            in nix.nix(["build", f"path:{subdir}"], cwd=subdir).run().expect(1).stderr_s
        )

    def test_flake_search_goes_up_without_installable(self, nix: Nix):
        nix.nix(["build", "--override-input", "foo", "."]).run().ok()

    def test_flake_inputs_do_not_search(self, nix: Nix, files: Path):
        flake = files / "flake.nix"
        flake.write_text(flake.read_text().replace(str(files / "foo"), str(files / "foo/subdir")))
        nix.nix(["build"]).run().expect(1)

    @pytest.mark.parametrize("args", good_uris)
    def test_flake_search_does_not_cross_git_repo(
        self, git: Git, nix: Nix, files: Path, args: list[str]
    ):
        subdir = files / "subdir"
        git(subdir, "init")
        assert (
            "is not part of a flake"
            in nix.nix(["build", *[arg.replace("$PWD", str(subdir)) for arg in args]], cwd=subdir)
            .run()
            .expect(1)
            .stderr_s
        )

    def test_flake_search_does_not_cross_git_repo_with_path_ru(
        self, git: Git, nix: Nix, files: Path
    ):
        subdir = files / "subdir"
        git(subdir, "init")
        assert (
            "does not contain a '/flake.nix' file"
            in nix.nix(["build", f"path:{subdir}"], cwd=subdir).run().expect(1).stderr_s
        )



"""
See "The Purely Functional Software Deployment Model", fig. 5.2 [1].

[1]: E. Dolstra, “The purely functional software deployment model,” Ph.D., Universeit Utrecht, Utrecht, NL, 2006. [Online]. Available: https://edolstra.github.io/pubs/phd-thesis.pdf
"""

from abc import ABCMeta, abstractmethod
import dataclasses
import struct
from typing import Protocol, ClassVar


class Writable(Protocol):
    """Realistically could just be IOBase but this is more constrained"""

    def write(self, data: bytes, /) -> int: ...


@dataclasses.dataclass
class NarListener:
    data: Writable

    def literal(self, data: bytes):
        self.data.write(data)

    def int_(self, v: int):
        self.literal(struct.pack("<Q", v))

    def add_pad(self, data_len: int):
        npad = 8 - data_len % 8
        if npad == 8:
            npad = 0
        # FIXME(Jade): implement nonzero padding
        self.literal(b"\0" * npad)

    def str_(self, data: bytes):
        self.int_(len(data))
        self.literal(data)
        self.add_pad(len(data))


class NarItem(metaclass=ABCMeta):
    type_: bytes

    def serialize(self, out: NarListener):
        out.str_(b"type")
        out.str_(self.type_)
        self.serialize_type(out)

    @abstractmethod
    def serialize_type(self, out: NarListener):
        pass


@dataclasses.dataclass
class Regular(NarItem):
    executable: bool
    contents: bytes
    type_: ClassVar[bytes] = b"regular"

    def serialize_type(self, out: NarListener):
        if self.executable:
            out.str_(b"executable")
            out.str_(b"")
        out.str_(b"contents")
        out.str_(self.contents)


@dataclasses.dataclass
class DirectoryUnordered(NarItem):
    entries: list[tuple[bytes, NarItem]]
    """Entries in the directory, not required to be in order because this nar is evil"""
    type_: ClassVar[bytes] = b"directory"

    @staticmethod
    def entry(out: NarListener, name: bytes, item: "NarItem"):
        # lol this format
        out.str_(b"entry")
        out.str_(b"(")
        out.str_(b"name")
        out.str_(name)
        out.str_(b"node")
        out.str_(b"(")
        item.serialize(out)
        out.str_(b")")
        out.str_(b")")

    def serialize_type(self, out: NarListener):
        for name, entry in self.entries:
            self.entry(out, name, entry)


@dataclasses.dataclass
class Directory(NarItem):
    entries: dict[bytes, NarItem]

    def serialize_type(self, out: NarListener):
        for name, item in sorted(self.entries.items(), key=lambda v: v[0]):
            DirectoryUnordered.entry(out, name, item)


@dataclasses.dataclass
class Symlink(NarItem):
    target: bytes
    type_: ClassVar[bytes] = b"symlink"

    def serialize_type(self, out: NarListener):
        out.str_(b"target")
        out.str_(self.target)


def serialize_nar(toplevel: NarItem, out: NarListener):
    out.str_(b"nix-archive-1")
    out.str_(b"(")
    toplevel.serialize(out)
    out.str_(b")")


def write_with_export_header(nar: NarItem, name: bytes, out: NarListener):
    # n.b. this is *not* actually a nar serialization, it just happens that nix
    # used exactly the same format for ints and strings in its protocol (and
    # nix-store --export) as it did in NARs lol
    export_magic = 0x4558494E

    # Store::exportPaths
    # For each path, put 1 then exportPath
    out.int_(1)

    # Store::exportPath
    serialize_nar(nar, out)
    out.int_(export_magic)
    # Due to `nix` setting the store using chroots /nix/store is still the test-local store and not the global one
    store_path = b"/nix/store"
    out.str_(store_path + b"/aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa-" + name)
    # no references
    out.int_(0)
    # no deriver
    out.str_(b"")
    # end of path
    out.int_(0)

    out.int_(0)



import dataclasses
from enum import Enum
from textwrap import dedent, indent
from typing import NamedTuple


from common import cxx_literal, generate_file, load_data, get_argument_parser

IMPURE_NOTE = """
> **Note**
>
> Not available in [pure evaluation mode](@docroot@/command-ref/conf-file.md#conf-pure-eval).

"""


class TypeName(NamedTuple):
    human: str
    code: str


class BuiltinType(TypeName, Enum):
    attrs = TypeName("set", "nAttrs")
    boolean = TypeName("boolean", "nBool")
    integer = TypeName("integer", "nInt")
    list = TypeName("list", "nList")
    null = TypeName("null", "nNull")
    string = TypeName("string", "nString")

    @classmethod
    def from_string(cls, t_name: str) -> "BuiltinType":
        for t in cls:
            if t_name == t.name:
                return t
        msg = f"Invalid builtin type: {t_name}"
        raise ValueError(msg)


@dataclasses.dataclass
class BuiltinConstant:
    name: str
    documentation: str

    # Fields with different name in the Post than in here
    # our fields
    type: BuiltinType = dataclasses.field(init=False)

    # Post fields
    type_str: dataclasses.InitVar[str]
    constructor_args: dataclasses.InitVar[list[str] | None] = None

    implementation: str = ""
    impure: bool = False
    rename_in_global_scope: bool = True

    def __post_init__(self, type_str: str, constructor_args: list[str] | None):
        self.type = BuiltinType.from_string(type_str)
        if constructor_args is not None:
            args = [f"NewValueAs::{type_str}"] + constructor_args
            self.implementation = f"{{{','.join(args)}}}"

    @property
    def code(self) -> str:
        cond = "if (!evalSettings.pureEval) " if self.impure else ""
        return dedent(f"""
            {cond} {{
                addConstant(
                    {cxx_literal(("__" if self.rename_in_global_scope else "") + self.name)},
                    {self.implementation},
                    {{
                        .type = {self.type.code},
                        .doc = {cxx_literal(self.documentation)},
                        .impureOnly = {cxx_literal(self.impure)},
                    }}
                );
            }}
        """)

    @property
    def docs(self) -> str:
        indentation = "    " * 3
        return dedent(f"""
            <dt id="builtins-{self.name}">
              <a href="#builtins-{self.name}"><code>{self.name}</code></a> ({self.type.human})
            </dt>
            <dd>

            {indent(self.documentation, indentation)}
            {indent(IMPURE_NOTE, indentation) if self.impure else ""}

            </dd>

        """)


def main():
    args = get_argument_parser().parse_args()

    builtin_constants = load_data(args.defs, BuiltinConstant)

    generate_file(
        args.header,
        builtin_constants,
        lambda constant:
        # `builtins` is magic and must come first
        "" if constant.name == "builtins" else constant.name,
        lambda b: b.code,
    )
    generate_file(args.docs, builtin_constants, lambda constant: constant.name, lambda b: b.docs)


if __name__ == "__main__":
    main()



from testlib.terminal_code_eater import eat_terminal_codes


def test_eats_color():
    assert eat_terminal_codes(b"\x1b[7mfoo blah bar\x1b[0m") == b"foo blah bar"


def test_eats_osc():
    assert (
        eat_terminal_codes(b"\x1b]8;;http://example.com\x1b\\This is a link\x1b]8;;\x1b\\")
        == b"This is a link"
    )



import dataclasses
from textwrap import dedent
from typing import Any

from common import (
    cxx_literal,
    generate_file,
    load_data,
    get_experimental_features,
    get_argument_parser,
)


PLATFORM_WARNING = """
  > **Note**
  > This setting is only available on {platforms} systems.

"""

XP_WARNING = """
  > **Warning**
  > This setting is part of an
  > [experimental feature](@docroot@/contributing/experimental-features.md).

  To change this setting, you need to make sure the corresponding experimental feature,
  [`{feature}`](@docroot@/contributing/experimental-features.md#xp-feature-{feature}),
  is enabled.
  For example, include the following in [`nix.conf`](#):

  ```
  extra-experimental-features = {feature}
  {name} = ...
  ```

"""

DEPR_WARNING = """
  > **Warning**
  > This setting is deprecated and will be removed in a future version of Lix.

"""


@dataclasses.dataclass
class Setting:
    name: str
    internal_name: str
    documentation: str

    default_text: str = ""
    setting_type: str = ""
    default_expr: str = ""
    platforms: list[str] = dataclasses.field(default_factory=list)
    aliases: list[str] = dataclasses.field(default_factory=list)
    experimental_feature: str | None = None
    deprecated: bool = False

    default: dataclasses.InitVar[str | None] = None
    type_str: dataclasses.InitVar[str | None] = None

    def __post_init__(self, default: Any, type_str: str | None):
        if default is not None:  # is not None nor an empty String
            self.default_text = f"`{nix_conf_literal(default)}`"
        self.default_expr = self.default_expr or cxx_literal(default)
        self.default_text = self.default_text or "*empty*"

        if type_str is not None:
            self.setting_type = f"Setting<{type_str}>"

    def generate_code(self, experimental_features: dict[str | None, str]) -> str:
        indentation = "    " * 4
        expr = (indent(indentation, self.default_expr) + indentation) if "\n" in self.default_expr else self.default_expr
        return dedent(f"""
            {self.setting_type} {self.internal_name} {{
                this,
                {expr},
                {cxx_literal(self.name)},
                {cxx_literal(self.documentation)},
                {cxx_literal(self.aliases)},
                true,
                {experimental_features[self.experimental_feature]},
                {cxx_literal(self.deprecated)}
            }};
        """)

    @property
    def docs(self) -> str:
        indentation = "    " * 3
        platforms = [p.capitalize() for p in self.platforms]
        aliases = [f"`{item}`" for item in self.aliases]
        description = dedent(f"""

{indent(indentation, self.documentation)}

            {indent(indentation, PLATFORM_WARNING.format(platforms=str(platforms)[1:-1])) if self.platforms else ""}
            {indent(indentation, XP_WARNING.format(feature=self.experimental_feature, name=self.name)) if self.experimental_feature is not None else ""}
            {indent(indentation, DEPR_WARNING) if self.deprecated else ""}
            **Default:** {self.default_text}
            {f"**Deprecated alias:** {str(aliases)[1:-1]}\n" if self.aliases else ""}
        """)
        return f'- <span id="conf-{self.name}">[`{self.name}`](#conf-{self.name})</span>' + indent(
            "  ",  # indent by two space to make it part of the list point
            description,
        )


platform_names = {"darwin": "Darwin", "linux": "Linux"}


def nix_conf_literal(v: Any) -> str:
    if v is None:
        return ""
    if v is False:
        return "false"
    if v is True:
        return "true"
    if isinstance(v, int):
        return str(v)
    if isinstance(v, str):
        return v
    if isinstance(v, list):
        return " ".join([nix_conf_literal(item) for item in v])
    msg = f"Cannot represent {v!r} in nix.conf"
    raise NotImplementedError(msg)


def indent(prefix: str, body: str) -> str:
    return "".join(["\n" if not line else f"{prefix}{line}\n" for line in body.split("\n")])


def main():
    ap = get_argument_parser()
    ap.add_argument("--kernel", help="Name of the kernel Lix will run on")
    ap.add_argument(
        "--experimental-features", help="Directory containing the experimental feature definitions"
    )
    args = ap.parse_args()

    settings = load_data(args.defs, Setting)

    experimental_features = get_experimental_features(
        args.experimental_features, [s.experimental_feature for (_, s) in settings]
    )

    generate_file(
        args.header,
        settings,
        lambda setting: setting.name,
        lambda setting: setting.generate_code(experimental_features)
        if not setting.platforms or args.kernel in setting.platforms
        else "",
    )
    generate_file(args.docs, settings, lambda setting: setting.name, lambda setting: setting.docs)


if __name__ == "__main__":
    main()



import re
import sys

from testlib.fixtures.file_helper import with_files, CopyFile
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset_pack

_files = {
    "output-cycles.nix": CopyFile("assets/test_build/output-cycles.nix"),
    **get_global_asset_pack("dependencies"),
}


def _assert_cycle_tree(output: str, regexes: list[str]):
    lines = output.splitlines()
    start = next((k for k, v in enumerate(lines) if "Shown below are the files inside" in v), None)

    assert start is not None

    for line, regex in enumerate(regexes, start=start + 1):
        assert re.search(regex, lines[line])


def _assert_cycle_message(err: str):
    assert (
        len(
            re.findall(
                r"cycle detected in build of '.*' in the references of output 'bar' from output 'foo'",
                err,
            )
        )
        == 1
    )


@with_files(_files)
def test_cycle(nix: Nix):
    res = nix.nix_build(["output-cycles.nix", "-A", "cycle"]).run().expect(1)
    err = res.stderr_plain
    _assert_cycle_message(err)

    if sys.platform == "linux":
        _assert_cycle_tree(
            err,
            [
                r"/store/.*-cycle-bar",
                r"└───lib/libfoo: ….*cycle-baz.*",
                r"    →.*/store/.*-cycle-baz",
                r"    └───share/lalala:.*-cycle-foo.*",
            ],
        )


@with_files(_files)
def test_cycle_in_dependency(nix: Nix):
    res = nix.nix_build(["output-cycles.nix", "-A", "as_dependency"]).run().expect(1)
    err = res.stderr_plain
    _assert_cycle_message(err)

    assert "error: 1 dependencies of derivation" in err


@with_files(_files)
def test_cycle_with_deps(nix: Nix):
    res = nix.nix_build(["output-cycles.nix", "-A", "cycle-with-deps"]).run().expect(1)
    err = res.stderr_plain
    _assert_cycle_message(err)

    if sys.platform == "linux":
        _assert_cycle_tree(
            err,
            [
                r"/store/.*-cycle-with-deps-bar",
                r"└───txt: ….*cycle-with-deps-foo.*",
                r"    →.*/store/.*-cycle-with-deps-foo",
                r"    └───txt:.*-cycle-with-deps-bar.*",
            ],
        )



from pathlib import Path

import pytest

from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.nix import NixSettings


def test_nix_settings_set_item():
    settings = NixSettings()
    settings["hello"] = "world"

    assert settings._settings["hello"] == "world"


def test_nix_settings_get_item():
    settings = NixSettings()
    assert settings["substituters"] == []


def test_nix_settings_set_attr():
    settings = NixSettings()
    settings.cores = 5

    assert settings._settings["cores"] == 5


def test_nix_settings_get_attr():
    settings = NixSettings()
    assert settings.sandbox is True


def test_nix_settings_get_attr_underscore():
    settings = NixSettings()
    assert settings.extra_deprecated_features == []

    settings["extra-deprecated-features"] += ["ancient-let"]

    assert settings.extra_deprecated_features == ["ancient-let"]


def test_nix_settings_set_attr_underscore():
    settings = NixSettings()
    assert settings["extra-deprecated-features"] == []

    settings.extra_deprecated_features += ["ancient-let"]
    assert settings["extra-deprecated-features"] == ["ancient-let"]


def test_nix_settings_update_replaces():
    settings = NixSettings()
    assert settings.sandbox is True

    settings.update({"sandbox": False})
    assert settings.sandbox is False

    settings.update(sandbox=True)
    assert settings.sandbox is True


def test_nix_settings_with_doesnt_effect_orig():
    orig = NixSettings()
    orig["extra-experimental-features"] += ["nix-command"]

    new = orig.with_settings({"extra-experimental-features": ["some-feature"]})
    assert new["extra-experimental-features"] == ["some-feature"]
    assert orig["extra-experimental-features"] == ["nix-command"]


def test_nix_settings_serializes_xf(env: ManagedEnv):
    settings = NixSettings()
    settings["extra-experimental-features"] += ["a", "b"]

    assert "extra-experimental-features = a b\n" in settings.to_config(env)


def test_nix_settings_serializes_store(env: ManagedEnv):
    settings = NixSettings()
    settings.store = "local?root=/some/path"

    assert "store = local?root=/some/path\n" in settings.to_config(env)


def test_nix_settings_serializes_both(env: ManagedEnv):
    settings = NixSettings()
    settings["extra-experimental-features"] += ["a", "b"]
    settings.store = "local?root=/some/path"

    serialized = settings.to_config(env)

    assert "extra-experimental-features = a b\n" in serialized
    assert "store = local?root=/some/path" in serialized


def test_nix_settings_ser_fails_bad_top_level_type(env: ManagedEnv):
    settings = NixSettings()
    settings.experimental_features = {"a": "b"}  # type: ignore we are testing the types here

    with pytest.raises(ValueError, match=r"Value is unsupported in nix config: {'a': 'b'}"):
        settings.to_config(env)


def test_nix_settings_ser_fails_bad_sub_type(env: ManagedEnv):
    settings = NixSettings()
    settings.experimental_features = [["a", "b"], "c"]  # type: ignore we are testing the types here

    with pytest.raises(ValueError, match=r"Value is unsupported in nix config: .+"):
        settings.to_config(env)


def test_nix_settings_to_env_overlay_no_store_dir(tmp_path: Path):
    env = ManagedEnv(tmp_path)
    settings = NixSettings()
    settings.store = "local?root=/some/path"

    settings.to_env_overlay(env)
    assert "store = local?root=/some/path\n" in env._env["NIX_CONFIG"]



from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import CopyFile, with_files

files = {"case.nar": CopyFile("assets/case.nar")}

opts = ["--option", "use-case-hack", "true"]

expected = {
    "x",
    "x/FOO",
    "x/Foo~nix~case~hack~1",
    "x/foo~nix~case~hack~2",
    "x/foo~nix~case~hack~2/A",
    "x/foo~nix~case~hack~2/A/foo",
    "x/foo~nix~case~hack~2/a~nix~case~hack~1",
    "x/foo~nix~case~hack~2/a~nix~case~hack~1/foo",
    "x/foo-1",
    "x/foo0",
    "xt_CONNMARK.h",
    "xt_CONNmark.h~nix~case~hack~1",
    "xt_connmark.h~nix~case~hack~2",
}


@with_files(files)
def test_case_hack(nix: Nix):
    nar_path = nix.env.dirs.home / "case.nar"
    nar = nar_path.read_bytes()
    out = nix.env.dirs.home / "case"

    # Check whether restoring and dumping a NAR that contains case
    # collisions is round-tripping, even on a case-insensitive system.
    nix.nix_store([*opts, "--restore", out]).with_stdin(nar).run().ok()
    assert set({str(p.relative_to(out)) for p in out.rglob("*")}) == expected

    dumped = nix.nix_store([*opts, "--dump", out]).run().ok().stdout
    assert dumped == nar

    out_hash = nix.nix([*opts, "--type", "sha256", out], nix_exe="nix-hash").run().ok().stdout_plain
    nar_hash = (
        nix.nix(["--flat", "--type", "sha256", nar_path], nix_exe="nix-hash")
        .run()
        .ok()
        .stdout_plain
    )
    assert out_hash == nar_hash

    # Check whether we detect true collisions (e.g. those remaining after
    # removal of the suffix).
    (out / "xt_CONNMARK.h~nix~case~hack~3").write_text("")
    assert "file name collision" in nix.nix_store([*opts, "--dump", out]).run().expect(1).stderr_s



#!/usr/bin/env python3
# Deletes the PCH arguments from a compilation database, to workaround nixpkgs
# stdenv having a cc-wrapper that is impossible to use for anything except cc
# itself, for example, clang-tidy.

import json
import shlex
import argparse


def process_compdb(compdb: list[dict]) -> list[dict]:
    def munch_command(args: list[str]) -> list[str]:
        out = []
        eat_next = False
        for i, arg in enumerate(args):
            if arg in ["-fpch-preprocess", "-fpch-instantiate-templates"]:
                # -fpch-preprocess as used with gcc, -fpch-instantiate-templates as used by clang
                continue
            if arg == "-include-pch" or (
                arg == "-include" and args[i + 1] == "precompiled-headers.hh"
            ):
                # -include-pch some-pch (clang), or -include some-pch (gcc)
                eat_next = True
                continue
            if not eat_next:
                out.append(arg)
            eat_next = False
        return out

    def chomp(item: dict) -> dict:
        item = item.copy()
        item["command"] = shlex.join(munch_command(shlex.split(item["command"])))
        return item

    def cmdfilter(item: dict) -> bool:
        file = item["file"]
        return not file.endswith("precompiled-headers.hh") and not file.endswith(".rs")

    return [chomp(x) for x in compdb if cmdfilter(x)]


def main():
    ap = argparse.ArgumentParser(description="Delete pch arguments from compilation database")
    ap.add_argument("input", type=argparse.FileType("r"), help="Input json file")
    ap.add_argument("output", type=argparse.FileType("w"), help="Output json file")
    args = ap.parse_args()

    input_json = json.load(args.input)
    json.dump(process_compdb(input_json), args.output, indent=2)


if __name__ == "__main__":
    main()



from pathlib import Path
from testlib.fixtures.file_helper import with_files
from testlib.utils import get_global_asset
from testlib.fixtures.nix import Nix


@with_files({"config.nix": get_global_asset("config.nix")})
def test_placeholders(nix: Nix):
    expr = """
      with import ./config.nix;

      mkDerivation {
        name = "placeholders";
        outputs = [ "out" "bin" "dev" ];
        buildCommand = "
          echo foo1 > $out
          echo foo2 > $bin
          echo foo3 > $dev
          [[ $(cat ${placeholder "out"}) = foo1 ]]
          [[ $(cat ${placeholder "bin"}) = foo2 ]]
          [[ $(cat ${placeholder "dev"}) = foo3 ]]
        ";
      }
    """

    result = Path(nix.nix_build(["-E", expr]).run().ok().stdout_plain)
    assert result.read_text() == "foo1\n"



import dataclasses
import shutil
import subprocess
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from collections.abc import Callable

import pytest
from testlib.fixtures.env import ManagedEnv
from testlib.fixtures.nix import Nix
from testlib.fixtures.command import Command


@pytest.fixture(autouse=True)
def common_init(nix: Nix):
    nix.settings.add_xp_feature("nix-command")
    nix.env.path.add_program("hg")


@dataclass
class HgRepo:
    path: Path
    rev1: str
    rev2: str

    @property
    def uri(self) -> str:
        # Intentionally not in a canonical form
        # See https://github.com/NixOS/nix/issues/6195
        return f'"file://{self.path.parent}/./{self.path.name}"'


@pytest.fixture(scope="session")
def static_hg_repo(tmp_path_factory: pytest.TempPathFactory) -> HgRepo:
    """
    create a static mercurial repo with our default settings because hg is *so slow*
    """
    repo = tmp_path_factory.mktemp("hg")

    subprocess.check_call(["hg", "init", repo])
    (repo / ".hg/hgrc").write_text(
        dedent("""
        [ui]
        username = Foobar <foobar@example.org>

        # Set ui.tweakdefaults to ensure HGPLAIN is being set.
        tweakdefaults = True
    """)
    )

    (repo / "hello").write_text("utrecht")
    (repo / ".hgignore").touch()
    subprocess.check_call(["hg", "add", "--cwd", repo, "hello", ".hgignore"])
    subprocess.check_call(["hg", "commit", "--cwd", repo, "-m", "Bla1"])
    rev1 = (
        subprocess.run(
            ["hg", "log", "--cwd", repo, "-r", "tip", "--template", "{node}"],
            stdout=subprocess.PIPE,
            check=True,
        )
        .stdout.strip()
        .decode("utf8", errors="replace")
    )

    (repo / "hello").write_text("world")
    subprocess.check_call(["hg", "commit", "--cwd", repo, "-m", "Bla2"])
    rev2 = (
        subprocess.run(
            ["hg", "log", "--cwd", repo, "-r", "tip", "--template", "{node}"],
            stdout=subprocess.PIPE,
            check=True,
        )
        .stdout.strip()
        .decode("utf8", errors="replace")
    )

    subprocess.check_call(
        ["hg", "--cwd", repo, "bookmark", "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"]
    )

    return HgRepo(repo, rev1, rev2)


@pytest.fixture
def hg(static_hg_repo: HgRepo, env: ManagedEnv) -> HgRepo:
    repo = env.dirs.home / "hg"
    shutil.copytree(static_hg_repo.path, repo)
    return dataclasses.replace(static_hg_repo, path=repo)


def test_fetch_unclean(hg: HgRepo, nix: Nix):
    (hg.path / "hello").write_text("unclean")
    result = (
        nix.nix(
            ["eval", "--impure", "--raw", "--expr", f"(builtins.fetchMercurial {hg.uri}).outPath"]
        )
        .run()
        .ok()
    )
    assert "is unclean" in result.stderr_plain
    assert (Path(result.stdout_plain) / "hello").read_text() == "unclean"


def test_fetch_current_rev(hg: HgRepo, nix: Nix):
    expr = f"(builtins.fetchMercurial {hg.uri}).outPath"
    result = nix.nix(["eval", "--impure", "--raw", "--expr", expr]).run().ok()
    assert (Path(result.stdout_plain) / "hello").read_text() == "world"


def test_pure_fetch_without_rev(hg: HgRepo, nix: Nix):
    expr = f'builtins.readFile (builtins.fetchMercurial {hg.uri} + "/hello")'
    result = nix.nix(["eval", "--raw", "--expr", expr]).run().expect(1)
    assert (
        "error: in pure evaluation mode, 'fetchMercurial' requires a Mercurial revision"
        in result.stderr_s
    )


def test_pure_fetch_with_good_rev(hg: HgRepo, nix: Nix):
    expr = f'(builtins.fetchMercurial {{ url = {hg.uri}; rev = "{hg.rev2}"; }}).outPath'
    result = nix.nix(["eval", "--raw", "--expr", expr]).run().ok()
    assert (Path(result.stdout_plain) / "hello").read_text() == "world"


def test_pure_fetch_with_bad_rev(hg: HgRepo, nix: Nix):
    expr = f'(builtins.fetchMercurial {{ url = {hg.uri}; rev = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"; }}).outPath'
    result = nix.nix(["eval", "--raw", "--expr", expr]).run().expect(1)
    assert "hg failed with exit code" in result.stderr_s


def test_fetch_attrs(hg: HgRepo, nix: Nix):
    expr = f'builtins.removeAttrs (builtins.fetchMercurial {{ url = {hg.uri}; rev = "{hg.rev2}"; }}) [ "outPath" ]'
    result = nix.nix(["eval", "--impure", "--json", "--expr", expr]).run().json()
    assert result["branch"] == "default"
    assert result["revCount"] == 2
    assert result["rev"] == hg.rev2


def test_caching(hg: HgRepo, nix: Nix):
    expr = f"(builtins.fetchMercurial {hg.uri}).outPath"
    path1 = nix.nix(["eval", "--impure", "--raw", "--expr", expr]).run().ok().stdout_plain
    shutil.rmtree(hg.path)
    path2 = nix.nix(["eval", "--impure", "--raw", "--expr", expr]).run().ok().stdout_plain
    assert path1 == path2
    assert (Path(path2) / "hello").read_text() == "world"

    # with TTL 0 refetch should fail
    result = nix.nix(["eval", "--impure", "--refresh", "--expr", expr]).run().expect(1)
    assert "error: 'hg pull' failed" in result.stderr_s


class TestExplicitHashes:
    def test_rev2(self, hg: HgRepo, nix: Nix):
        expr = f'(builtins.fetchMercurial {{ url = {hg.uri}; rev = "{hg.rev2}"; }}).outPath'
        path1 = nix.nix(["eval", "--raw", "--expr", expr]).run().ok().stdout_plain
        shutil.rmtree(hg.path)
        path2 = nix.nix(["eval", "--refresh", "--raw", "--expr", expr]).run().ok().stdout_plain
        assert path1 == path2

    def test_rev1(self, hg: HgRepo, nix: Nix):
        expr = f'(builtins.fetchMercurial {{ url = {hg.uri}; rev = "{hg.rev1}"; }}).outPath'
        path2 = nix.nix(["eval", "--refresh", "--raw", "--expr", expr]).run().ok().stdout_plain
        assert (Path(path2) / "hello").read_text() == "utrecht"


class TestFetchDirt:
    @pytest.fixture(scope="class")
    def static_changed_repo(
        self, static_hg_repo: HgRepo, tmp_path_factory: pytest.TempPathFactory
    ) -> HgRepo:
        repo = tmp_path_factory.mktemp("hg-changed")
        shutil.copytree(static_hg_repo.path, repo, dirs_exist_ok=True)

        for name in ["dir1", "dir2"]:
            (repo / name).mkdir()
        (repo / "dir1/foo").write_text("foo")
        (repo / "bar").write_text("bar")
        (repo / "dir2/bar").write_text("bar")
        subprocess.check_call(["hg", "add", "--cwd", repo, "dir1/foo"])
        subprocess.check_call(["hg", "rm", "--cwd", repo, "hello"])

        return dataclasses.replace(static_hg_repo, path=repo)

    # this intentionally overrides the module-level fixture name
    @pytest.fixture
    def hg(self, static_changed_repo: HgRepo, env: ManagedEnv) -> HgRepo:
        repo = env.dirs.home / "hg"
        shutil.copytree(static_changed_repo.path, repo)
        return dataclasses.replace(static_changed_repo, path=repo)

    def test_fetch_simple(self, hg: HgRepo, nix: Nix, command: Callable[..., Command]):
        expr = f"""
            let f = builtins.fetchMercurial {hg.uri};
            in {{ path = f.outPath; rev = f.rev; }}
        """
        result = nix.nix(["eval", "--impure", "--json", "--expr", expr]).run().json()
        path = Path(result["path"])
        assert not (path / "hello").exists()
        assert not (path / "bar").exists()
        assert not (path / "dir2/bar").exists()
        assert (path / "dir1/foo").read_text() == "foo"
        assert result["rev"] == "0000000000000000000000000000000000000000"

        command(["hg", "commit", "--cwd", hg.path, "-m", "Bla3"]).run().ok()
        result2 = nix.nix(["eval", "--impure", "--json", "--expr", expr]).run().json()
        assert result["path"] == result2["path"]

    def test_fetch_rev(self, hg: HgRepo, nix: Nix):
        expr = f"""
            let f = builtins.fetchMercurial {{ url = {hg.uri}; rev = "default"; }};
            in {{ path = f.outPath; rev = f.rev; }}
        """
        result = nix.nix(["eval", "--impure", "--json", "--expr", expr]).run().json()
        path = Path(result["path"])
        assert (path / "hello").exists()
        assert not (path / "bar").exists()
        assert not (path / "dir2/bar").exists()
        assert not (path / "dir1/foo").exists()
        assert result["rev"] == hg.rev2


def test_fetch_with_name(hg: HgRepo, nix: Nix):
    """Passing a `name` argument should be reflected in the output path"""
    (hg.path / "hello").write_text("paris")
    expr = f'(builtins.fetchMercurial {{ url = {hg.uri}; name = "foo"; }}).outPath'
    path = (
        nix.nix(["eval", "--impure", "--refresh", "--raw", "--expr", expr]).run().ok().stdout_plain
    )
    assert path.endswith("-foo")


class TestNondefaultBranch:
    @pytest.fixture(autouse=True)
    def setup(self, hg: HgRepo, command: Callable[..., Command]):
        (hg.path / "hello").write_text("paris")
        command(["hg", "branch", "--cwd", hg.path, "meow"]).run().ok()
        command(["hg", "commit", "--cwd", hg.path, "-m", "Bla3"]).run().ok()

    def test_simple(self, hg: HgRepo, nix: Nix):
        expr = f'builtins.removeAttrs (builtins.fetchMercurial {hg.uri}) [ "outPath" ]'
        result = nix.nix(["eval", "--impure", "--json", "--expr", expr]).run().json()
        assert result["branch"] == "meow"
        assert result["revCount"] == 3

    def test_older_rev(self, hg: HgRepo, nix: Nix, command: Callable[..., Command]):
        command(["hg", "update", "--cwd", hg.path, "1"]).run().ok()
        expr = f'builtins.removeAttrs (builtins.fetchMercurial {hg.uri}) [ "outPath" ]'
        result = nix.nix(["eval", "--impure", "--json", "--expr", expr]).run().json()
        assert result["branch"] == "default"
        assert result["revCount"] == 2



from testlib.fixtures.nix import Nix


def test_trivial_addition(nix: Nix):
    assert nix.eval("1 + 1").json() == 2


def test_trivial_assert(nix: Nix):
    assert nix.eval("assert 1 + 2 == 3; true").json() is True



from pathlib import Path
from collections.abc import Callable

import pytest

from testlib.fixtures.command import CommandResult, Command, RunningCommand
from testlib.fixtures.env import ManagedEnv


type GitCmd = Callable[[Path | None, *tuple[str | Path, ...]], RunningCommand]
type Git = Callable[[Path | None, *tuple[str | Path, ...]], CommandResult]


@pytest.fixture
def git_cmd(env: ManagedEnv) -> GitCmd:
    """
    Provides a convenient wrapper to call git. Adds `git` to PATH, sets author and commit
    date environment variables to fixed values. If `path` is not None it will be added to
    the git invocation a `-C` argument. The remaining arguments are forwarded to Command,
    then `run` is called on the resulting Command object. The RunningCommand is returned.
    """

    env.path.add_program("git")
    # ensure that commit hashes are as deterministic as possible
    env["GIT_AUTHOR_DATE"] = "@42 +0000"
    env["GIT_COMMITTER_DATE"] = "@23 +0000"

    def call(path: Path | None, *args: str | Path, **kwargs: str | Path) -> RunningCommand:
        return Command(
            ["git", *(["-C", path] if path is not None else []), *args], _env=env, **kwargs
        ).run()

    return call


@pytest.fixture
def git(git_cmd: GitCmd) -> Git:
    """
    Provides a convenient wrapper to call git. Same as `git_cmd`, but also calls `ok()`.
    """

    def call(path: Path | None, *args: str | Path, **kwargs: str | Path) -> CommandResult:
        return git_cmd(path, *args, **kwargs).ok()

    return call



import pytest
from testlib.fixtures.nix import Nix
from testlib.fixtures.file_helper import File, with_files


class TestImpureEval:
    @pytest.fixture(autouse=True)
    def setup(self, nix: Nix):
        nix.settings.add_xp_feature("nix-command")
        self.nix = nix

    def eval(self, expr: str, *args: str) -> str:
        return (
            self.nix.nix(["eval", "--impure", "--raw", *args, "--expr", expr])
            .run()
            .ok()
            .stdout_plain
        )

    @pytest.mark.parametrize("path", ["/foo", "/bar"])
    def test_store_dir(self, path: str):
        result = self.eval("builtins.storeDir", "--store", f"dummy://?store={path}")
        assert result == path

    @pytest.mark.parametrize("system", ["foo", "bar"])
    def test_current_system_system_only(self, system: str):
        result = self.eval("builtins.currentSystem", "--system", system)
        assert result == system

    @pytest.mark.parametrize("system", ["foo", "bar"])
    def test_current_system_both(self, system: str):
        result = self.eval("builtins.currentSystem", "--system", system, "--eval-system", system)
        assert result == system

    @pytest.mark.parametrize("system", ["foo", "bar"])
    def test_current_system_eval_system_only(self, system: str):
        result = self.eval("builtins.currentSystem", "--eval-system", system)
        assert result == system

    @pytest.mark.parametrize("system", ["foo", "bar"])
    def test_current_system_both_override(self, system: str):
        result = self.eval("builtins.currentSystem", "--system", system, "--eval-system", "baz")
        assert result == "baz"

    @with_files({"eval.nix": File('{ attr.foo = "bar"; }')})
    def test_nix_path_search(self, nix: Nix):
        result = (
            nix.nix(
                [
                    "eval",
                    "--option",
                    "nix-path",
                    f"foo={nix.env.dirs.home}",
                    "-f",
                    "<foo/eval.nix>",
                    "attr.foo",
                ]
            )
            .run()
            .ok()
        )
        assert result.stdout_plain == '"bar"'



import os
import stat
from pathlib import Path

from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset


@with_files({"config.nix": get_global_asset("config.nix")})
def test_build_dir_permissions(nix: Nix):
    """
    ensure that the build directory parent is not world-accessible
    """

    build_dir = nix.env.dirs.home / "build-dir"
    build_dir.mkdir(0o755)
    fifo = build_dir / "fifo"
    os.mkfifo(fifo)
    expr = f"""
        with import ./config.nix; mkDerivation {{
            name = "test";
            buildCommand = "echo >'{fifo}'; cat '{fifo}' > $out";
        }}
    """

    nix.settings.add_xp_feature("nix-command")
    nix.settings.extra_sandbox_paths = [str(fifo)]

    build = nix.nix(["build", "--build-dir", f"{build_dir}/b", "-E", expr, "--impure"]).start()

    fifo.read_text()
    try:
        child_dir = Path(next(iter((build_dir / "b").iterdir())))
        assert stat.S_IMODE(child_dir.stat().st_mode) in [0o700, 0o710]
    finally:
        fifo.write_text("")

    build.wait().ok()



from string import Template

import pytest


def test_template_balanced_kwargs(balanced_templater: type[Template]):
    assert balanced_templater("@test_str@").substitute(test_str="valid") == "valid"


def test_template_balanced_multi_use(balanced_templater: type[Template]):
    assert balanced_templater("@test_str@ @test_str@").substitute(test_str="valid") == "valid valid"


def test_balanced_escapement(balanced_templater: type[Template]):
    assert balanced_templater("@@ @@{test} @{test}@").substitute(test="t") == "@ @{test} t"


# And that correct exceptions get thrown
def test_template_balanced_extra_kwarg(balanced_templater: type[Template]):
    with pytest.raises(ExceptionGroup) as excinfo:
        balanced_templater("{test_str}").substitute(test_str="valid", random_str="random")
    assert excinfo.group_contains(KeyError, match=r"random_str")
    assert not excinfo.group_contains(ValueError)


def test_template_balanced_missing_arg(balanced_templater: type[Template]):
    with pytest.raises(KeyError) as e:
        balanced_templater("@TEST@ @MISSING@").substitute(TEST="valid")
    assert e.match("MISSING")



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2001, 2002, 2003, 2004, 2005, 2006 Python Software Foundation; All Rights Reserved
# SPDX-FileCopyrightText: 2024 Jade Lovelace
# SPDX-License-Identifier: LGPL-2.1-or-later
"""
This script exists to lose Lix a dependency on expect(1) for the ability to run
something in a pty.

Yes, it could be replaced by script(1) but macOS and Linux script(1) have
diverged sufficiently badly that even specifying a subcommand to run is not the
same.
"""
import pty
import sys
import os
from termios import ONLCR, ONLRET, ONOCR, OPOST, TCSAFLUSH, tcgetattr, tcsetattr
from tty import setraw
import termios

def setup_terminal():
    # does not matter which fd we use because we are in a fresh pty
    modi = tcgetattr(pty.STDOUT_FILENO)
    [iflag, oflag, cflag, lflag, ispeed, ospeed, cc] = modi

    # Turning \n into \r\n is not cool, Linux!
    oflag &= ~ONLCR
    # I don't know what "implementation dependent postprocessing means" but it
    # sounds bad
    oflag &= ~OPOST
    # Assume that NL performs the role of CR; do not insert CRs at column 0
    oflag |= ONLRET | ONOCR

    modi = [iflag, oflag, cflag, lflag, ispeed, ospeed, cc]

    tcsetattr(pty.STDOUT_FILENO, TCSAFLUSH, modi)


def spawn(argv: list[str]):
    """
    As opposed to pty.spawn, this one more seriously controls the pty settings.
    Necessary to turn off such fun functionality as onlcr (LF to CRLF).

    This is essentially copy pasted from pty.spawn, since there is no way to
    hook the child pre-execve
    """
    pid, master_fd = pty.fork()
    if pid == pty.CHILD:
        setup_terminal()
        os.execlp(argv[0], *argv)

    try:
        mode = tcgetattr(pty.STDIN_FILENO)
        setraw(pty.STDIN_FILENO)
        restore = True
    except termios.error:
        restore = False

    try:
        pty._copy(master_fd, pty._read, pty._read) # type: ignore
    finally:
        if restore:
            tcsetattr(pty.STDIN_FILENO, TCSAFLUSH, mode) # type: ignore

    os.close(master_fd)
    return os.waitpid(pid, 0)[1]


def main():
    if len(sys.argv) == 1:
        print(f'Usage: {sys.argv[0]} [command args]', file=sys.stderr)
        sys.exit(1)

    sys.exit(os.waitstatus_to_exitcode(spawn(sys.argv[1:])))


if __name__ == '__main__':
    main()



from pathlib import Path

from testlib.fixtures.file_helper import with_files
from testlib.fixtures.nix import Nix
from testlib.utils import get_global_asset


@with_files({"config.nix": get_global_asset("config.nix")})
class TestOptimizeStore:
    def _test_optimise_store(self, nix: Nix):
        def build(name: str, *extra_args: str) -> Path:
            expr = f"""
                with import ./config.nix; mkDerivation {{
                    name = "{name}";
                    builder = builtins.toFile "builder" "mkdir $out; echo hello > $out/foo";
                }}
            """

            result = nix.nix_build(["-E", expr, "--no-out-link", *extra_args]).run().ok()
            return Path(result.stdout_plain)

        out1 = build("foo1", "--auto-optimise-store")
        out2 = build("foo2", "--auto-optimise-store")

        assert (out1 / "foo").samefile(out2 / "foo")
        assert (out1 / "foo").stat().st_nlink == 3

        out3 = build("foo3", "--no-auto-optimise-store")

        assert not (out1 / "foo").samefile(out3 / "foo")

        nix.nix_store(["--optimise"]).run().ok()

        assert (out1 / "foo").samefile(out3 / "foo")

        nix.nix_store(["--gc"]).run().ok()

        assert list((nix.env.dirs.real_store_dir / ".links").glob("*")) == []

    def test_optimise_store(self, nix: Nix):
        self._test_optimise_store(nix)

    def test_optimise_store_daemon(self, nix: Nix):
        nix.settings.auto_optimise_store = True
        with nix.daemon([], {"trusted-users": "*"}) as inner:
            self._test_optimise_store(inner)



from testlib.fixtures.nix import Nix
from testlib.fixtures.git import Git
from testlib.fixtures.file_helper import with_files, File
from testlib.utils import get_global_asset_pack
from pathlib import Path


@with_files(
    {
        "flakeA": get_global_asset_pack(".git")
        | {
            "flake.nix": File("""{
          inputs.b.url = "git+file://$flakeB";
          inputs.b.inputs.a.follows = "/";

          outputs = { self, b }: {
            foo = 123 + b.bar;
            xyzzy = 1000;
          };
        }""")
        },
        "flakeB": get_global_asset_pack(".git")
        | {
            "flake.nix": File("""{
          inputs.a.url = "git+file://$flakeA";

          outputs = { self, a }: {
            bar = 456 + a.xyzzy;
          };
        }""")
        },
    }
)
def test_circular(nix: Nix, files: Path, git: Git):
    nix.settings.add_xp_feature("nix-command", "flakes")

    path = files / "flakeA/flake.nix"
    path.write_text(path.read_text().replace("$flakeB", str(files / "flakeB")))
    git(files / "flakeA", "add", "flake.nix")

    path = files / "flakeB/flake.nix"
    path.write_text(path.read_text().replace("$flakeA", str(files / "flakeA")))
    git(files / "flakeB", "add", "flake.nix")
    git(files / "flakeB", "commit", "-a", "-m", "Foo")

    assert nix.nix(["eval", f"{files}/flakeA#foo"]).run().ok().stdout_s == "1579\n"

    path = files / "flakeB/flake.nix"
    path.write_text(path.read_text().replace("456", "789"))
    git(files / "flakeB", "commit", "-a", "-m", "Foo")
    nix.nix(["flake", "update", "b", "--flake", f"{files}/flakeA"]).run().ok()
    assert nix.nix(["eval", f"{files}/flakeA#foo"]).run().ok().stdout_s == "1912\n"

    # Test list-inputs with circular dependencies
    nix.nix(["flake", "metadata", f"{files}/flakeA"]).run().ok()
